package io.testcase;

import io.testcase.model.AnalyticalNominalOutput;
import io.testcase.model.Anomaly;
import io.testcase.model.Argument;
import io.testcase.model.CartesianParameter;
import io.testcase.model.Coordinate;
import io.testcase.model.EccentricityVector;
import io.testcase.model.EquinoctialParameter;
import io.testcase.model.Geodetic;
import io.testcase.model.InclinationVector;
import io.testcase.model.KeplerianParameter;
import io.testcase.model.OrbitOrientation;
import io.testcase.model.OrbitShapeSize;
import io.testcase.model.Other;
import io.testcase.model.SubSatellitePoint;
import lombok.Data;
import lombok.RequiredArgsConstructor;
import org.hipparchus.geometry.euclidean.threed.Vector3D;
import org.hipparchus.util.FastMath;
import org.orekit.bodies.GeodeticPoint;
import org.orekit.bodies.OneAxisEllipsoid;
import org.orekit.frames.Frame;
import org.orekit.orbits.KeplerianOrbit;
import org.orekit.orbits.OrbitType;
import org.orekit.propagation.SpacecraftState;
import org.orekit.propagation.sampling.OrekitFixedStepHandler;
import org.orekit.time.AbsoluteDate;
import org.orekit.time.TimeScalesFactory;

import java.util.ArrayList;
import java.util.List;
import java.util.Objects;

import static org.orekit.utils.Constants.WGS84_EARTH_EQUATORIAL_RADIUS;

@Data
@RequiredArgsConstructor
public class EphemerisHandler implements OrekitFixedStepHandler {

  private List<AnalyticalNominalOutput> outputs = new ArrayList<>();

  private final OneAxisEllipsoid earth;

  private final Frame satelliteFrame;

  private final AbsoluteDate startDate;

  @Override
  public void init(SpacecraftState s0, AbsoluteDate t, double step) {
  }

  @Override
  public void handleStep(SpacecraftState spacecraftState, boolean b) {
    KeplerianOrbit keplerianOrbit = (KeplerianOrbit) OrbitType.KEPLERIAN.convertType(spacecraftState.getOrbit());

    outputs.add(AnalyticalNominalOutput.builder()
        .date(spacecraftState.getDate().toDate(TimeScalesFactory.getUTC()).toInstant())
        .cartesianParameter(getCartesianParameter(spacecraftState))
        .keplerianParameter(getKeplerianParameter(keplerianOrbit))
        .equinoctialParameter(getEquinoctialParameter(keplerianOrbit))
        .other(getOther(spacecraftState))
        .subSatellitePoint(getSubSatellitePoint(spacecraftState))
        .build());
  }

  private CartesianParameter getCartesianParameter(SpacecraftState state) {
    Vector3D orekitPosition = state.getPVCoordinates().getPosition();
    Coordinate position = new Coordinate(orekitPosition.getX(), orekitPosition.getY(), orekitPosition.getZ());
    Vector3D orekitVelocity = state.getPVCoordinates().getVelocity();
    Coordinate velocity = new Coordinate(orekitVelocity.getX(), orekitVelocity.getY(), orekitVelocity.getZ());
    Vector3D orekitAcceleration = state.getPVCoordinates().getAcceleration();
    Coordinate acceleration = new Coordinate(orekitAcceleration.getX(), orekitAcceleration.getY(), orekitAcceleration.getZ());

    return new CartesianParameter(position, velocity, acceleration);
  }

  private OrbitShapeSize getOrbitShapeSize(KeplerianOrbit keplerianOrbit) {
    double ra = keplerianOrbit.getA() * (1 + keplerianOrbit.getE());
    double rp = keplerianOrbit.getA() * (1 - keplerianOrbit.getE());

    double ha = ra - WGS84_EARTH_EQUATORIAL_RADIUS;
    double hp = rp - WGS84_EARTH_EQUATORIAL_RADIUS;

    return new OrbitShapeSize(ha, hp, ra, rp, keplerianOrbit.getA(), keplerianOrbit.getE());
  }

  private static Double radiansToDegrees(Double radian, int maxDegree) {
    if (Objects.isNull(radian)) {
      return radian;
    }
    Double degrees = 0d;
    if (radian < 0) {
      degrees = maxDegree * 1d;
    }
    return degrees + FastMath.toDegrees(radian);
  }

  private KeplerianParameter getKeplerianParameter(KeplerianOrbit keplerianOrbit) {
    OrbitOrientation orbitOrientation = new OrbitOrientation(radiansToDegrees(keplerianOrbit.getI(), 180),
        radiansToDegrees(keplerianOrbit.getRightAscensionOfAscendingNode(), 360),
        radiansToDegrees(keplerianOrbit.getPerigeeArgument(), 360));

    Anomaly anomaly = new Anomaly(radiansToDegrees(keplerianOrbit.getTrueAnomaly(), 360),
        radiansToDegrees(keplerianOrbit.getEccentricAnomaly(), 360),
        radiansToDegrees(keplerianOrbit.getMeanAnomaly(), 360));


    return new KeplerianParameter(getOrbitShapeSize(keplerianOrbit), orbitOrientation, anomaly);
  }

  private EquinoctialParameter getEquinoctialParameter(KeplerianOrbit keplerianOrbit) {
    EccentricityVector eccentricityVector = new EccentricityVector(keplerianOrbit.getEquinoctialEx(), keplerianOrbit.getEquinoctialEy());

    InclinationVector inclinationVector = new InclinationVector(keplerianOrbit.getHx(), keplerianOrbit.getHy());

    Argument argument = new Argument(keplerianOrbit.getLE(), keplerianOrbit.getLM(), keplerianOrbit.getLv());


    return new EquinoctialParameter(keplerianOrbit.getA(), eccentricityVector, inclinationVector, argument);
  }

  private Other getOther(SpacecraftState state) {
    Vector3D orekitAngular = state.getPVCoordinates().getAngularVelocity();
    Coordinate angularVelocity = new Coordinate(orekitAngular.getX(), orekitAngular.getY(), orekitAngular.getZ());
    Vector3D orekitMomentum = state.getPVCoordinates().getMomentum();
    Coordinate momentum = new Coordinate(orekitMomentum.getX(), orekitMomentum.getY(), orekitMomentum.getZ());
    return new Other(state.getKeplerianPeriod(), state.getKeplerianMeanMotion(), angularVelocity, momentum);
  }

  private SubSatellitePoint getSubSatellitePoint(SpacecraftState state) {
    GeodeticPoint satelliteSubPoint = earth.transform(state.getPVCoordinates().getPosition(), satelliteFrame, state.getDate());
    Geodetic geodetic = new Geodetic(satelliteSubPoint.getLatitude(), satelliteSubPoint.getLongitude(), satelliteSubPoint.getAltitude());

    Vector3D cartesianVector = earth.transform(
        new GeodeticPoint(FastMath.toRadians(satelliteSubPoint.getLatitude()), FastMath.toRadians(satelliteSubPoint.getLongitude()),
            FastMath.toRadians(satelliteSubPoint.getAltitude())));

    Coordinate cartesian = new Coordinate(cartesianVector.getX(), cartesianVector.getY(), cartesianVector.getZ());

    return new SubSatellitePoint(geodetic, cartesian);
  }
}
