package io.testcase;

import io.testcase.model.AnalyticalNominalInput;
import io.testcase.model.AnalyticalNominalOutput;
import io.testcase.model.AnalyticalNominalPropagating;
import lombok.SneakyThrows;
import lombok.extern.slf4j.Slf4j;
import org.orekit.attitudes.AttitudeProvider;
import org.orekit.attitudes.NadirPointing;
import org.orekit.bodies.OneAxisEllipsoid;
import org.orekit.data.DataContext;
import org.orekit.data.DataProvidersManager;
import org.orekit.data.DirectoryCrawler;
import org.orekit.frames.Frame;
import org.orekit.frames.FramesFactory;
import org.orekit.propagation.analytical.tle.TLE;
import org.orekit.propagation.analytical.tle.TLEPropagator;
import org.orekit.time.AbsoluteDate;
import org.orekit.time.TimeScale;
import org.orekit.time.TimeScalesFactory;
import org.orekit.utils.IERSConventions;

import java.io.File;
import java.net.URL;
import java.time.Instant;
import java.util.Date;
import java.util.List;
import java.util.Objects;

import static org.orekit.utils.Constants.WGS84_EARTH_EQUATORIAL_RADIUS;
import static org.orekit.utils.Constants.WGS84_EARTH_FLATTENING;

@Slf4j
public class TestCase implements AnalyticalNominalPropagating {

  public static final String NAME = "AnalyticalNominalPropagator";

  @SneakyThrows
  public void readFile() {
    // TODO: get data in resource folder orekit-data
    URL resource = getClass().getClassLoader().getResource("orekit-data");
    File orekitData = new File(resource.toURI());
    DataProvidersManager manager = DataContext.getDefault().getDataProvidersManager();
    manager.addProvider(new DirectoryCrawler(orekitData));
  }

  @SneakyThrows
  @Override
  public List<AnalyticalNominalOutput> propagate(AnalyticalNominalInput input) {
    readFile();
    TimeScale timescale = TimeScalesFactory.getUTC();

    Frame earthFrame = FramesFactory.getITRF(IERSConventions.IERS_2010, true);
    OneAxisEllipsoid earth = new OneAxisEllipsoid(
        WGS84_EARTH_EQUATORIAL_RADIUS,
        WGS84_EARTH_FLATTENING,
        earthFrame
    );
    Frame satelliteFrame = FramesFactory.getGCRF();
    AttitudeProvider attitudeProvider = new NadirPointing(satelliteFrame, earth);
    org.orekit.propagation.Propagator propagator = TLEPropagator.selectExtrapolator(
        new TLE(input.getTle1(), input.getTle2()),
        attitudeProvider,
        input.getMass(),
        satelliteFrame
    );
    Date startDate = Date.from(input.getStartDate());
    AbsoluteDate absoluteStartDate = new AbsoluteDate(startDate, timescale);
    AbsoluteDate absoluteEndDate = new AbsoluteDate(absoluteStartDate, input.getDuration());
    Integer step = Objects.nonNull(input.getEphemerisTimestep()) ? input.getEphemerisTimestep() : input.getDuration();
    EphemerisHandler ephemerisHandler = new EphemerisHandler(earth, satelliteFrame, absoluteStartDate);
    propagator.setMasterMode(step, ephemerisHandler);
    propagator.propagate(absoluteStartDate, absoluteEndDate);
    return ephemerisHandler.getOutputs();
  }

  public static void main(String[] args) {
    AnalyticalNominalInput input = AnalyticalNominalInput.builder()
        .satelliteName("testSatellite")
        .startDate(Instant.parse("2020-04-06T05:00:00.000Z"))
        .duration(3600)
        .tle1("1 25544U 98067A   20097.18686503  .00000920  00000-0  25115-4 0  9994")
        .tle2("2 25544  51.6465 344.7546 0003971  92.6495  47.0504 15.48684294220855")
        .mass(31100)
        .ephemerisTimestep(60).build();

    List<AnalyticalNominalOutput> output = new TestCase().propagate(input);
  }
}
